# This is a very simple Jinja2 template representing an imaginary configuration file
# for an imaginary app.

# this is an example of loading a fact from the setup module
system={{ ansible_system }}

# here is a variable that could be set in a playbook or inventory file
http_port={{ http_port }}


